; Digital I/O peripheral for SC/MP computer
;
; K Orton 2017

; PIC constants
W	EQU	0
PCL	EQU	H'0002'
STATUS	EQU	H'0003'
PORTA	EQU	H'0005'
PORTB	EQU	H'0006'
PIR1	EQU	H'000C'
RCSTA	EQU	H'0018'
TXREG	EQU	H'0019'
RCREG	EQU	H'001A'
CMCON	EQU	H'001F'
TRISA	EQU	H'0085'
TRISB	EQU	H'0086'
TXSTA	EQU	H'0098'
SPBRG	EQU	H'0099'
C	EQU	0
Z	EQU	2
RP0	EQU	5
CFG	EQU	H'3F01'

	__CONFIG	CFG

; Program constants
TMP1	EQU	H'0070'
TMP2	EQU	H'0071'
TMP3	EQU	H'0072'
DIN1	EQU	0
DIN2	EQU	1
DIN3	EQU	2
DIN4	EQU	3
LED	EQU	4
CTS	EQU	0
DOUT1	EQU	4
DOUT2	EQU	5
DOUT3	EQU	6
DOUT4	EQU	7
CRCH	EQU	H'0D'

; Reset
	ORG	0
	MOVLW	7
	MOVWF	CMCON
	MOVLW	B'00010000'
	MOVWF	PORTA
	CLRF	PORTB
	BSF	STATUS,RP0
	MOVLW	B'00101111'
	MOVWF	TRISA
	MOVLW	B'00000111'
	MOVWF	TRISB
	MOVLW	B'00100100'
	MOVWF	TXSTA
	MOVLW	D'207'
	MOVWF	SPBRG
	BCF	STATUS,RP0
	BSF	RCSTA,7
	NOP
	BSF	RCSTA,4

; Main program loop
WTCMD	BSF	PORTA,LED
L1	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSS	STATUS,C
	GOTO	L1
	ADDLW	D'256'+' '-'Z'-1
	BTFSC	STATUS,C
	GOTO	ABORT
	ADDLW	'Z'+1-'A'
	BTFSS	STATUS,C
	GOTO	ABORT
	BCF	PORTA,LED
	ADDWF	PCL
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	SETDO
	GOTO	INTDIO
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
ABORT	BSF	PORTA,LED
L2	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSC	STATUS,C
	GOTO	L2
	GOTO	L1

; Get character from computer
; Character returned in W
GETCH	BTFSS	PIR1,5
	GOTO	GETCH
	MOVF	RCREG,W
	BTFSC	RCSTA,1
	BCF	RCSTA,4
	NOP
	BSF	RCSTA,4
	RETURN

; Send character back to computer
; Character expected in W
PUTCH	BTFSS	PIR1,4
	GOTO	PUTCH
	BTFSC	PORTB,CTS
	GOTO	PUTCH
	MOVWF	TXREG
	RETURN

; Get number from computer
; Number returned in W in range 0..99
; Aborts if an error occurs
; STATUS.Z set if zero returned
; Uses TMP2, TMP3
GETNUM	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSS	STATUS,C
	GOTO	ABORT
	BTFSC	STATUS,Z
	GOTO	GETNUM
	ADDLW	D'256'+' '-'9'-1
	BTFSC	STATUS,C
	GOTO	ABORT
	ADDLW	'9'+1-'0'
	BTFSS	STATUS,C
	GOTO	ABORT
	MOVWF	TMP2
	CALL	GETCH
	ADDLW	D'256'-' '-1
	BTFSS	STATUS,C
	GOTO	L3
	ADDLW	D'256'+' '-'9'
	BTFSC	STATUS,C
	GOTO	ABORT
	ADDLW	'9'+1-'0'
	BTFSS	STATUS,C
	GOTO	ABORT
	MOVWF	TMP3
	CALL	GETCH
	ADDLW	D'256'-' '-1
	BTFSC	STATUS,C
	GOTO	ABORT
	RLF	TMP2,W
	ANDLW	B'00011110'
	ADDWF	TMP3
	RLF	TMP2
	RLF	TMP2
	RLF	TMP2,W
	ANDLW	B'01111000'
	ADDWF	TMP3,W
	RETURN
L3	MOVF	TMP2,W
	RETURN

; Send number back to computer
; Number expected in W in range 0..99
; Uses TMP2, TMP3
PUTNUM	MOVWF	TMP2
	MOVLW	-1
	MOVWF	TMP3
	MOVLW	D'10'
L4	INCF	TMP3
	SUBWF	TMP2
	BTFSC	STATUS,C
	GOTO	L4
	MOVF	TMP3,W
	BTFSC	STATUS,Z
	GOTO	L5
	ADDLW	'0'
	CALL	PUTCH
L5	MOVF	TMP2,W
	ADDLW	D'10'+'0'
	CALL	PUTCH
	RETURN

; Set digital outputs
; Uses TMP1
SETDO	CLRF	TMP1
	CALL	GETNUM
	BTFSS	STATUS,Z
	BSF	TMP1,DOUT1
	CALL	GETNUM
	BTFSS	STATUS,Z
	BSF	TMP1,DOUT2
	CALL	GETNUM
	BTFSS	STATUS,Z
	BSF	TMP1,DOUT3
	CALL	GETNUM
	BTFSS	STATUS,Z
	BSF	TMP1,DOUT4
	MOVF	TMP1,W
	MOVWF	PORTB
	GOTO	WTCMD

; Interrogate digital I/O
INTDIO	MOVLW	0
	BTFSC	PORTB,DOUT1
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSC	PORTB,DOUT2
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSC	PORTB,DOUT3
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSC	PORTB,DOUT4
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSS	PORTA,DIN1
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSS	PORTA,DIN2
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSS	PORTA,DIN3
	MOVLW	1
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVLW	0
	BTFSS	PORTA,DIN4
	MOVLW	1
	CALL	PUTNUM
	MOVLW	CRCH
	CALL	PUTCH
	GOTO	WTCMD

	END
