; Remote control receiver, time of day clock and
;  non-volatile memory peripheral for SC/MP computer
;
; K Orton 2017

; PIC constants
W	EQU	0
PCL	EQU	H'0002'
STATUS	EQU	H'0003'
PORTA	EQU	H'0005'
PORTB	EQU	H'0006'
PIR1	EQU	H'000C'
TMR1L	EQU	H'000E'
TMR1H	EQU	H'000F'
T1CON	EQU	H'0010'
CCPR1L	EQU	H'0015'
CCPR1H	EQU	H'0016'
CCP1CN	EQU	H'0017'
RCSTA	EQU	H'0018'
TXREG	EQU	H'0019'
RCREG	EQU	H'001A'
CMCON	EQU	H'001F'
TRISB	EQU	H'0086'
TXSTA	EQU	H'0098'
SPBRG	EQU	H'0099'
EEDATA	EQU	H'009A'
EEADR	EQU	H'009B'
EECON1	EQU	H'009C'
EECON2	EQU	H'009D'
C	EQU	0
Z	EQU	2
RP0	EQU	5
RD	EQU	0
WR	EQU	1
WREN	EQU	2
CFG	EQU	H'3F01'

	__CONFIG	CFG

; Program constants
TICKS	EQU	H'0070'
SECS	EQU	H'0071'
MINS	EQU	H'0072'
HRS	EQU	H'0073'
CSECS	EQU	H'0074'
CMINS	EQU	H'0075'
CHRS	EQU	H'0076'
PRVKEY	EQU	H'0077'
NEWKEY	EQU	H'0078'
TMP1	EQU	H'0079'
TMP2	EQU	H'007A'
TMP3	EQU	H'007B'
KEYMSK	EQU	H'0F'
LB	EQU	4
RTS	EQU	0
CTS	EQU	4
LED	EQU	5
CRCH	EQU	H'0D'

; Reset
	ORG	0
	MOVLW	7
	MOVWF	CMCON
	MOVLW	B'00000001'
	MOVWF	PORTB
	BSF	STATUS,RP0
	MOVLW	B'00010110'
	MOVWF	TRISB
	MOVLW	B'00100100'
	MOVWF	TXSTA
	MOVLW	D'170'
	MOVWF	SPBRG
	BCF	STATUS,RP0
	MOVLW	HIGH(D'51199')
	MOVWF	CCPR1H
	MOVLW	LOW(D'51199')
	MOVWF	CCPR1L
	CLRF	TMR1H
	CLRF	TMR1L
	MOVLW	B'00001011'
	MOVWF	CCP1CN
	BSF	RCSTA,7

; Program initialisation
	CLRF	TICKS
	CLRF	SECS
	CLRF	MINS
	CLRF	HRS
	CLRF	PRVKEY
	CLRF	NEWKEY
	BSF	RCSTA,4
	BSF	T1CON,0

; Main program loop
WTCMD	BCF	PORTB,LED
L1	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSS	STATUS,C
	GOTO	L1
	ADDLW	D'256'+' '-'Z'-1
	BTFSC	STATUS,C
	GOTO	ABORT
	ADDLW	'Z'+1-'A'
	BTFSS	STATUS,C
	GOTO	ABORT
	BSF	PORTB,LED
	ADDWF	PCL
	GOTO	ABORT
	GOTO	INTREM
	GOTO	SETTOD
	GOTO	INTTOD
	GOTO	SETNVM
	GOTO	INTNVM
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
	GOTO	ABORT
ABORT	BCF	PORTB,LED
L2	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSC	STATUS,C
	GOTO	L2
	GOTO	L1

; Get character from computer
; Character returned in W
; Services time of day clock while waiting
GETCH	CALL	SRVCLK
	BTFSS	PIR1,5
	GOTO	GETCH
	MOVF	RCREG,W
	BTFSC	RCSTA,1
	BCF	RCSTA,4
	NOP
	BSF	RCSTA,4
	RETURN

; Send character back to computer
; Character expected in W
; Services time of day clock while waiting
; Uses TMP1
PUTCH	MOVWF	TMP1
L3	CALL	SRVCLK
	BTFSS	PIR1,4
	GOTO	L3
	BTFSC	PORTB,CTS
	GOTO	L3
	MOVF	TMP1,W
	MOVWF	TXREG
	RETURN

; Get number from computer
; Number returned in W in range 0..99
; Aborts if an error occurs
; STATUS.Z set if zero returned
; Uses TMP2, TMP3
GETNUM	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSS	STATUS,C
	GOTO	ABORT
	BTFSC	STATUS,Z
	GOTO	GETNUM
	ADDLW	D'256'+' '-'9'-1
	BTFSC	STATUS,C
	GOTO	ABORT
	ADDLW	'9'+1-'0'
	BTFSS	STATUS,C
	GOTO	ABORT
	MOVWF	TMP2
	CALL	GETCH
	ADDLW	D'256'-' '-1
	BTFSS	STATUS,C
	GOTO	L4
	ADDLW	D'256'+' '-'9'
	BTFSC	STATUS,C
	GOTO	ABORT
	ADDLW	'9'+1-'0'
	BTFSS	STATUS,C
	GOTO	ABORT
	MOVWF	TMP3
	CALL	GETCH
	ADDLW	D'256'-' '-1
	BTFSC	STATUS,C
	GOTO	ABORT
	RLF	TMP2,W
	ANDLW	B'00011110'
	ADDWF	TMP3
	RLF	TMP2
	RLF	TMP2
	RLF	TMP2,W
	ANDLW	B'01111000'
	ADDWF	TMP3,W
	RETURN
L4	MOVF	TMP2,W
	RETURN

; Send number back to computer
; Number expected in W in range 0..99
; Uses TMP2, TMP3
PUTNUM	MOVWF	TMP2
	MOVLW	-1
	MOVWF	TMP3
	MOVLW	D'10'
L5	INCF	TMP3
	SUBWF	TMP2
	BTFSC	STATUS,C
	GOTO	L5
	MOVF	TMP3,W
	BTFSC	STATUS,Z
	GOTO	L6
	ADDLW	'0'
	CALL	PUTCH
L6	MOVF	TMP2,W
	ADDLW	D'10'+'0'
	CALL	PUTCH
	RETURN

; Reset non-volatile memory
RSTNVM	BSF	STATUS,RP0
	CLRF	EEADR
	BCF	STATUS,RP0
	RETURN

; Read non-volatile memory
; Non-volatile character returned in W
; STATUS.Z set if null character
RDNVM	BSF	STATUS,RP0
	BSF	EECON1,RD
	NOP
	INCF	EEADR
	MOVF	EEDATA,W
	BCF	STATUS,RP0
	RETURN

; Write non-volatile memory
; Character expected in W
; Services time of day clock while waiting
WRNVM	BCF	PORTB,RTS
	BSF	STATUS,RP0
	MOVWF	EEDATA
	BSF	EECON1,WREN
	MOVLW	H'55'
	MOVWF	EECON2
	MOVLW	H'AA'
	MOVWF	EECON2
	BSF	EECON1,WR
L7	BCF	STATUS,RP0
	CALL	SRVCLK
	BSF	STATUS,RP0
	BTFSC	EECON1,WR
	GOTO	L7
	BCF	EECON1,WREN
	INCF	EEADR
	BCF	STATUS,RP0
	BSF	PORTB,RTS
	RETURN

; Service clock
SRVCLK	BTFSS	PIR1,2
	RETURN
	BCF	PIR1,2
	COMF	PORTA,W
	ANDLW	KEYMSK
	BTFSC	STATUS,Z
	GOTO	L8
	COMF	PORTA,W
	ANDLW	KEYMSK
	BTFSC	STATUS,Z
	GOTO	L8
	COMF	PORTA,W
	ANDLW	KEYMSK
	BTFSC	STATUS,Z
	GOTO	L8
	MOVF	PRVKEY
	BTFSC	STATUS,Z
	MOVWF	NEWKEY
L8	MOVWF	PRVKEY
	MOVLW	D'256'/D'16'
	ADDWF	TICKS
	BTFSS	STATUS,C
	RETURN
	INCF	SECS
	MOVLW	D'256'-D'60'
	ADDWF	SECS,W
	BTFSS	STATUS,C
	RETURN
	CLRF	SECS
	INCF	MINS
	MOVLW	D'256'-D'60'
	ADDWF	MINS,W
	BTFSS	STATUS,C
	RETURN
	CLRF	MINS
	INCF	HRS
	MOVLW	D'256'-D'24'
	ADDWF	HRS,W
	BTFSC	STATUS,C
	CLRF	HRS
	RETURN

; Interrogate remote receiver key press
INTREM	MOVF	NEWKEY,W
	CLRF	NEWKEY
	CALL	PUTNUM
	MOVLW	CRCH
	CALL	PUTCH
	GOTO	WTCMD

; Set time of day
SETTOD	CALL	GETNUM
	MOVWF	CHRS
	CALL	GETNUM
	MOVWF	CMINS
	CALL	GETNUM
	MOVWF	SECS
	MOVF	CMINS,W
	MOVWF	MINS
	MOVF	CHRS,W
	MOVWF	HRS
	CLRF	TICKS
	GOTO	WTCMD

; Interrogate time of day
INTTOD	MOVF	SECS,W
	MOVWF	CSECS
	MOVF	MINS,W
	MOVWF	CMINS
	MOVF	HRS,W
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVF	CMINS,W
	CALL	PUTNUM
	MOVLW	','
	CALL	PUTCH
	MOVF	CSECS,W
	CALL	PUTNUM
	MOVLW	CRCH
	CALL	PUTCH
	GOTO	WTCMD

; Set non-volatile memory
SETNVM	CALL	RSTNVM
L9	CALL	GETCH
	ADDLW	D'256'-' '
	BTFSS	STATUS,C
	GOTO	L10
	ADDLW	' '
	CALL	WRNVM
	GOTO	L9
L10	MOVLW	0
	CALL	WRNVM
	GOTO	WTCMD

; Interrogate non-volatile memory
INTNVM	CALL	RSTNVM
L11	CALL	RDNVM
	BTFSC	STATUS,Z
	GOTO	L12
	CALL	PUTCH
	GOTO	L11
L12	MOVLW	CRCH
	CALL	PUTCH
	GOTO	WTCMD

	END
