; Speech synthesiser peripheral for SC/MP computer
;
; K Orton 2015
;
; Requires standard capacity SD card of 4Mbytes min

; PIC constants
W	EQU	0
INDF	EQU	H'0000'
PCL	EQU	H'0002'
STATUS	EQU	H'0003'
FSR	EQU	H'0004'
PORTA	EQU	H'0005'
PORTB	EQU	H'0006'
PCLATH	EQU	H'000A'
PIR1	EQU	H'000C'
T2CON	EQU	H'0012'
CCPR1L	EQU	H'0015'
CCP1CN	EQU	H'0017'
RCSTA	EQU	H'0018'
RCREG	EQU	H'001A'
CMCON	EQU	H'001F'
OPTREG	EQU	H'0081'
TRISA	EQU	H'0085'
TRISB	EQU	H'0086'
PR2	EQU	H'0092'
SPBRG	EQU	H'0099'
C	EQU	0
Z	EQU	2
RP0	EQU	5
IRP	EQU	7
CFG	EQU	H'3F02'

	__CONFIG	CFG

; Program constants
TMP1	EQU	H'000E'
TMP2	EQU	H'000F'
COMB	EQU	H'0020'
BLKCNT	EQU	H'007C'
BLKNO	EQU	H'007D'
ALLONO	EQU	H'007E'
SMPCNT	EQU	H'007F'
FIFIN	EQU	H'009A'
FIFOUT	EQU	H'009B'
FIFO	EQU	H'00A0'
HLFBUF	EQU	H'0120'
RSPCNT	EQU	H'014E'
BLKSZ	EQU	D'92'
FIFOSZ	EQU	D'80'
CRC	EQU	H'95'
SDMISO	EQU	0
SDSCLK	EQU	1
SDMOSI	EQU	2
SDCS	EQU	3
ADDFLG	EQU	5
SELFLG	EQU	6
LED	EQU	7
CMD	EQU	B'01000000'
SPCH	EQU	H'20'
MYADD	EQU	'A'

; Delay macro
; Uses W
DELAY	MACRO	CYCLES
	IF	((CYCLES>3)&&(CYCLES<8))
	CALL	DEL4
	ENDIF
	IF	(CYCLES>7)
	MOVLW	((CYCLES/4)-1)
	CALL	DEL4N
	ENDIF
	IF	((CYCLES-(4*(CYCLES/4)))>0)
	NOP
	ENDIF
	IF	((CYCLES-(4*(CYCLES/4)))>1)
	NOP
	ENDIF
	IF	((CYCLES-(4*(CYCLES/4)))>2)
	NOP
	ENDIF
	ENDM

; Reset
	ORG	0
	MOVLW	7
	MOVWF	CMCON
	MOVLW	B'00001100'
	MOVWF	PORTA
	CLRF	PORTB
	BSF	STATUS,RP0
	BCF	OPTREG,7
	MOVLW	B'00100001'
	MOVWF	TRISA
	MOVLW	B'00000110'
	MOVWF	TRISB
	MOVLW	D'155'
	MOVWF	SPBRG
	MOVLW	D'74'
	MOVWF	PR2
	BCF	STATUS,RP0
	MOVLW	B'10000000'
	MOVWF	RCSTA
	MOVLW	D'32'
	MOVWF	CCPR1L
	MOVLW	B'00001100'
	MOVWF	CCP1CN

; SD card initialisation
	MOVLW	D'80'
	MOVWF	TMP1
L1	BSF	PORTA,SDSCLK
	NOP
	NOP
	NOP
	BCF	PORTA,SDSCLK
	DECFSZ	TMP1
	GOTO	L1
	DELAY	D'120'
	CLRF	BLKCNT
	MOVLW	CMD+0
	CALL	SDCMD
L2	MOVLW	CMD+D'55'
	CALL	SDCMD
	MOVLW	CMD+D'41'
	CALL	SDCMD
	ANDLW	B'00000001'
	BTFSS	STATUS,Z
	GOTO	L2
	MOVLW	BLKSZ
	MOVWF	BLKCNT
	MOVLW	CMD+D'16'
	CALL	SDCMD

; Program initialisation
	MOVLW	COMB
	MOVWF	FSR
	MOVLW	BLKSZ
	MOVWF	TMP1
	MOVLW	D'128'
L3	MOVWF	INDF
	INCF	FSR
	DECFSZ	TMP1
	GOTO	L3
	BSF	STATUS,RP0
	MOVLW	(D'128'-FIFOSZ)
	MOVWF	FIFIN
	MOVWF	FIFOUT
	BCF	STATUS,RP0
	MOVLW	HLFBUF
	MOVWF	FSR
	BSF	STATUS,IRP
	MOVLW	(BLKSZ/2)
	MOVWF	TMP1
	MOVLW	D'128'
L4	MOVWF	INDF
	INCF	FSR
	DECFSZ	TMP1
	GOTO	L4
	BSF	RCSTA,4
	MOVLW	1
	MOVWF	BLKCNT
	CLRF	BLKNO
	CLRF	ALLONO
	MOVLW	HIGH(ALLSZT)
	MOVWF	PCLATH
	BSF	T2CON,2

; Main program loop
L5	BSF	PORTA,SDCS
	MOVLW	(BLKSZ/2)
	MOVWF	SMPCNT
	CALL	OPBSMP
	DECF	SMPCNT
	BCF	PORTA,SDCS
	MOVLW	CMD+D'17'
	CALL	SDSEND
	MOVLW	0
	CALL	SDSEND
	DELAY	D'59'
	CALL	OPBSMP
	DECF	SMPCNT
	MOVF	ALLONO,W
	CALL	SDSEND
	MOVF	BLKNO,W
	CALL	SDSEND
	DELAY	D'60'
	CALL	OPBSMP
	DECF	SMPCNT
	MOVLW	0
	CALL	SDSEND
	DELAY	D'166'
	CALL	OPBSMP
	DECF	SMPCNT
	MOVLW	CRC
	CALL	SDSEND
	MOVLW	RSPCNT
	MOVWF	FSR
	BSF	STATUS,IRP
	CLRF	INDF
	DELAY	D'19'
L6	DELAY	D'143'
	CALL	OPBSMP
	MOVLW	RSPCNT
	MOVWF	FSR
	BSF	STATUS,IRP
	BTFSC	INDF,1
	GOTO	L7
	CALL	SDRCV
	XORLW	H'FF'
	BTFSS	STATUS,Z
	INCF	INDF
	GOTO	L8
L7	DELAY	D'74'
L8	DELAY	D'47'
	DECFSZ	SMPCNT
	GOTO	L6
	MOVLW	(BLKSZ/2)
	MOVWF	SMPCNT
	CALL	OPSSMP
	DECF	SMPCNT
	INCF	BLKNO
	DECFSZ	BLKCNT
	GOTO	L10
	BSF	STATUS,RP0
	MOVF	FIFIN,W
	XORWF	FIFOUT,W
	BTFSS	STATUS,Z
	GOTO	L9
	BCF	STATUS,RP0
	INCF	BLKCNT
	CLRF	BLKNO
	CLRF	ALLONO
	DELAY	D'18'
	BCF	PORTB,LED
	GOTO	L11
L9	MOVF	FIFOUT,W
	ADDLW	(FIFO-(D'128'-FIFOSZ))
	MOVWF	FSR
	INCF	FIFOUT
	NOP
	MOVLW	(D'128'-FIFOSZ)
	BTFSC	FIFOUT,7
	MOVWF	FIFOUT
	BCF	STATUS,IRP
	BCF	STATUS,RP0
	MOVF	INDF,W
	ANDLW	B'00111111'
	MOVWF	ALLONO
	CLRF	BLKNO
	CALL	ALLSZT
	MOVWF	BLKCNT
	BSF	PORTB,LED
	GOTO	L11
L10	DELAY	D'29'
L11	CALL	POLLSP
	DELAY	D'15'
L12	DELAY	D'53'
	CALL	OPSSMP
	DELAY	D'75'
	DECFSZ	SMPCNT
	GOTO	L12
	CALL	SDRCV
	CALL	SDRCV
	CALL	POLLSP
	DELAY	D'22'
	GOTO	L5

; Send command to SD card (initialisation only)
; Command number expected in W
; Argument bits 7..0 expected in BLKCNT
; Single response byte returned in W
; Uses TMP1, TMP2, SMPCNT
SDCMD	BCF	PORTA,SDCS
	MOVWF	SMPCNT
	MOVLW	H'FF'
	CALL	SDSEND
	MOVF	SMPCNT,W
	CALL	SDSEND
	MOVLW	0
	CALL	SDSEND
	MOVLW	0
	CALL	SDSEND
	MOVLW	0
	CALL	SDSEND
	MOVF	BLKCNT,W
	CALL	SDSEND
	MOVLW	CRC
	CALL	SDSEND
L13	CALL	SDRCV
	XORLW	H'FF'
	BTFSC	STATUS,Z
	GOTO	L13
	XORLW	H'FF'
	BSF	PORTA,SDCS
	RETURN

; Send byte to SD card
; Leaves MOSI in high state
; Uses TMP1, TMP2
SDSEND	MOVWF	TMP1			; CALL SDSEND = 105~
	MOVLW	8
	MOVWF	TMP2
L14	BCF	PORTA,SDSCLK
	BTFSS	TMP1,7
	BCF	PORTA,SDMOSI
	BTFSC	TMP1,7
	BSF	PORTA,SDMOSI
	NOP
	NOP
	RLF	TMP1
	BSF	PORTA,SDSCLK
	DECFSZ	TMP2
	GOTO	L14
	NOP
	BCF	PORTA,SDSCLK
	BSF	PORTA,SDMOSI
	RETURN

; Receive byte from SD card
; Uses TMP1, TMP2
SDRCV	MOVLW	8			; CALL SDRCV = 70~
	MOVWF	TMP2
L15	BSF	PORTA,SDSCLK
	RRF	PORTA,W
	RLF	TMP1
	NOP
	BCF	PORTA,SDSCLK
	DECFSZ	TMP2
	GOTO	L15
	MOVF	TMP1,W
	RETURN

; Output buffered sample
; Sample index (BLKSZ/2)..1 expected in SMPCNT
; CCP1 update at +22~
; Uses TMP1, TMP2
OPBSMP	DECF	SMPCNT,W		; CALL OPBSMP = 27~
	ADDLW	HLFBUF
	MOVWF	FSR
	BSF	STATUS,IRP
	MOVF	INDF,W
	BCF	STATUS,IRP
	ADDWF	INDF
	RRF	INDF
	BTFSC	INDF,7
	BCF	STATUS,C
	BTFSC	STATUS,C
	INCF	INDF
	SWAPF	INDF,W
	ANDLW	B'00110000'
	IORLW	B'00001100'
	MOVWF	TMP1
	RRF	INDF,W
	MOVWF	TMP2
	RRF	TMP2,W
	ANDLW	B'00111111'
	MOVWF	CCPR1L
	MOVF	TMP1,W
	MOVWF	CCP1CN
	RETURN

; Output SD card sample (and buffer one)
; Sample index (BLKSZ/2)..1 expected in SMPCNT
; CCP1 update at +164~
; Uses TMP1, TMP2
OPSSMP	DECF	SMPCNT,W		; CALL OPSSMP = 169~
	ADDLW	HLFBUF
	MOVWF	FSR
	BSF	STATUS,IRP
	CALL	SDRCV
	MOVWF	INDF
	BCF	STATUS,IRP
	MOVLW	(BLKSZ/2)
	ADDWF	FSR
	CALL	SDRCV
	ADDWF	INDF
	RRF	INDF
	BTFSC	INDF,7
	BCF	STATUS,C
	BTFSC	STATUS,C
	INCF	INDF
	SWAPF	INDF,W
	ANDLW	B'00110000'
	IORLW	B'00001100'
	MOVWF	TMP1
	RRF	INDF,W
	MOVWF	TMP2
	RRF	TMP2,W
	ANDLW	B'00111111'
	MOVWF	CCPR1L
	MOVF	TMP1,W
	MOVWF	CCP1CN
	RETURN

; Poll serial port
; Uses TMP1
POLLSP	BTFSS	PIR1,5			; CALL POLLSP = 29~
	GOTO	L18
	MOVF	RCREG,W
	MOVWF	TMP1
	BTFSC	RCSTA,1
	BCF	RCSTA,4
	NOP
	BSF	RCSTA,4
	ADDLW	D'256'-SPCH
	BTFSS	STATUS,C
	GOTO	L16
	BTFSS	PORTB,SELFLG
	GOTO	L17
	BSF	STATUS,RP0
	MOVF	FIFIN,W
	ADDLW	(FIFO-(D'128'-FIFOSZ))
	MOVWF	FSR
	INCF	FIFIN
	MOVLW	(D'128'-FIFOSZ)
	BTFSC	FIFIN,7
	MOVWF	FIFIN
	BCF	STATUS,IRP
	BCF	STATUS,RP0
	MOVF	TMP1,W
	MOVWF	INDF
	RETURN
L16	BCF	PORTB,ADDFLG
	DELAY	D'11'
	BCF	PORTB,SELFLG
	RETURN
L17	BTFSC	PORTB,ADDFLG
	GOTO	L19
	BSF	PORTB,ADDFLG
	DELAY	5
	XORLW	MYADD-SPCH
	BTFSC	STATUS,Z
	BSF	PORTB,SELFLG
	RETURN
L18	DELAY	D'22'
	RETURN
L19	DELAY	8
	RETURN

; Delay routines
DEL4N	ADDLW	H'FF'			; CALL DEL4N = ((4*(W+1))-1)~
	BTFSS	STATUS,Z
	GOTO	DEL4N
DEL4	RETURN				; CALL DEL4 = 4~

; Allophone size table
ALLSZT	ADDWF	PCL			; CALL ALLSZT = 6~
	RETLW	D'1'
	RETLW	D'3'
	RETLW	D'5'
	RETLW	D'11'
	RETLW	D'22'
	RETLW	D'47'
	RETLW	D'29'
	RETLW	D'9'
	RETLW	D'12'
	RETLW	D'15'
	RETLW	D'17'
	RETLW	D'19'
	RETLW	D'9'
	RETLW	D'5'
	RETLW	D'20'
	RETLW	D'9'
	RETLW	D'21'
	RETLW	D'8'
	RETLW	D'27'
	RETLW	D'29'
	RETLW	D'32'
	RETLW	D'9'
	RETLW	D'12'
	RETLW	D'12'
	RETLW	D'12'
	RETLW	D'20'
	RETLW	D'14'
	RETLW	D'15'
	RETLW	D'10'
	RETLW	D'21'
	RETLW	D'12'
	RETLW	D'30'
	RETLW	D'42'
	RETLW	D'17'
	RETLW	D'11'
	RETLW	D'22'
	RETLW	D'8'
	RETLW	D'22'
	RETLW	D'22'
	RETLW	D'14'
	RETLW	D'18'
	RETLW	D'10'
	RETLW	D'14'
	RETLW	D'25'
	RETLW	D'25'
	RETLW	D'14'
	RETLW	D'21'
	RETLW	D'41'
	RETLW	D'24'
	RETLW	D'15'
	RETLW	D'19'
	RETLW	D'18'
	RETLW	D'34'
	RETLW	D'27'
	RETLW	D'33'
	RETLW	D'12'
	RETLW	D'22'
	RETLW	D'21'
	RETLW	D'37'
	RETLW	D'33'
	RETLW	D'39'
	RETLW	D'5'
	RETLW	D'22'
	RETLW	D'4'

	END
